from . import func


class BuildingGeometry:

    def __init__(self):
        """Step-1:
        Building geometry parameters 

        Properties:
            * story_height (tuple):  Story_height
            * bay_length (tuple): Bay_length
            * plan_dimensions (tuple): (main direction, secondary direction)  
            * MF_number (int): Number of moment frame
            * exterior_column_tributary_area (tuple): (x, y), area = x * y  
            * interior_column_tributary_area (tuple): (x, y), area = x * y  
        """
        self.story_height = None
        self.bay_length = None
        self.plane_dimensions = None
        self.exterior_column_tributary_area = None
        self.interior_column_tributary_area = None
        self.MF_number = None

    def _check(self):
        func.check_list(self.story_height, min_length=1, max_length=98, name='story')
        func.check_list(self.bay_length, min_length=1, max_length=98, name='story')
        func.check_tuple(self.plane_dimensions, 2, name='plan_dimensions')
        func.check_int(self.MF_number, [1, 99], name='MF_number')
        func.check_tuple(self.exterior_column_tributary_area, 2, name='exterior_column_tributary_area')
        func.check_tuple(self.interior_column_tributary_area, 2, name='exterior_column_tributary_area')

    def _finish(self):
        """Complete the definition of building geometry parameters"""
        self._check()
        self.N = len(self.story_height)
        self.bays = len(self.bay_length)
        self.axis = self.bays + 1
        self.building_height = sum(self.story_height)
        





from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .MRFhelper import Frame
from typing import Dict
from wsection import WSection, GBSection
from . import func


class StructuralComponents:

    def __init__(self, frame: Frame):
        """Step-2:
        Structural components parameters  
        """
        self.N = frame.N
        self.bays = frame.bays
        self.axis = frame.axis
        self.beams: Dict[int, str] = dict()  # Necessary
        self.columns: Dict[int, str] = dict()  # Necessary
        self.column_splice = tuple()  # Unecessary
        self.beam_splice = tuple()  # Unecessary
        self.RBS_length_all = None  # Use same RBS length for all beams
        self.doubler_plate = dict()  # Unecessary
        # initialize doubler plate thickness
        for floor in range(2, self.N + 2):
            self.doubler_plate[floor] = [0] * self.axis

    def set_beams(self, floor: int, sections: list[str]):
        """Set beam sizes

        Args:
            floor (int): floor number, must be larger than 2
            sections (list[str]): A list including all w-section types for the floor

        Example:
            >>> set_beams(2, ['W24x64', 'W24x64', 'W24x64'])
            The beam section for the 2rd floor is designated as W24x64
        """
        func.check_int(floor, min_max=[2, self.N + 1], name='floor')
        func.check_list(sections, length=self.bays, pos=False, name=' beam section')
        self.beams[floor] = sections


    def set_columns(self, story: int, sections: list[str]):
        """Set column sizes, if there is a column splice, input according to the
           cross-section at the bottom of the column

        Args:
            story (int): story number
            sections (list[str]): A list including all w-section types for the story
            
        """
        func.check_int(story, min_max=[1, self.N], name='axis')
        func.check_list(sections, length=self.axis, pos=False, name='column section')
        self.columns[story] = sections


    def set_column_splice(self, *story: int):
        """Set the stories where column splices are located.

        Args:
            story (list): Defaults to ().
        """
        if story:
            func.check_tuple(story, min_length=1, max_length=self.N)
            for val in story:
                func.check_int(val, [1, self.N], 'story')
                if val == self.N:
                    raise ValueError(f'Column splice cannot be placed at the top story ({self.N})')
            self.column_splice = story


    def set_beam_splice(self, *bay: int):
        """Set the stories where beam splices are located.

        Args:
            bay (list): Defaults to ().
        """
        if bay:
            func.check_tuple(bay, min_length=1, max_length=self.bays)
            for val in bay:
                func.check_int(val, [1, self.bays], 'bay')
            self.beam_splice = bay


    def set_doubler_plate(self, floor: int, thicknesses: list[int | float]):
        """(Optional) Set doubler plate (if any) thickness for the panel zone

        Args:
            floor (int): Floor number of the panel zone
            thicknesses (list[int  |  float]): A list including all doubler plate thickness values for the floor
        """
        func.check_int(floor, [2, 99], name='floor')
        func.check_list(thicknesses, length=self.axis, name='doubler plate thickness')
        self.doubler_plate[floor] = thicknesses


    def set_RBS_length(self, RBS_length: int | float):
        """(Optional) Set RBS length (distance from beam hinge to panel zone edge)

        Args:
            RBS_length (int | float): Offset distance of the beam hinge
        """
        func.check_int_float(RBS_length, name='RBS_length')
        self.RBS_length_all = RBS_length


    def _finish(self):
        floor_set = set([2 + i for i in range(self.N)])
        floor_defined_beam = set(self.beams.keys())
        if floor_set == floor_defined_beam:
            pass
        else:
            if floor_set - floor_defined_beam:
                raise ValueError(f'Beams on the floor {floor_set - floor_defined_beam} have not been defined')
            if floor_defined_beam - floor_set:
                raise ValueError(f'Beams on the floor {floor_defined_beam - floor_set} are not existed')
        for key, val in self.columns.items():
            if len(val) != self.axis:
                raise ValueError(f'Number of defined column of stort {key} is less than the axis number ({self.axis})')


    def _get_section_properties(self, frame: Frame):
        """Get beam and column properties
        * beam_properties (dict): {floor, [[bf, h, ...], [bf, h, ...], ...(x bays)]}
        * column_properties (dict): {story, [[bf, h, ...], [bf, h, ...], ...(x N)]}
        * RBS_length (dict): {floor: [l1, l2, ...(x 2*bays)]}
        """
        fy_beam = frame.LoadAndMaterial.fy_beam
        fy_column = frame.LoadAndMaterial.fy_column
        BC_connection = frame.ConnectionAndBoundary.beam_column_connection
        RBS_paras = frame.ConnectionAndBoundary.RBS_paras
        self.beam_properties = dict()
        for floor, sections in self.beams.items():
            props = []
            for section in sections:
                if section.startswith(('W', 'w')):
                    prop = WSection(section, fy_beam)
                elif section.startswith(('HW', 'HM', 'HN')):
                    prop = GBSection(section, fy_beam)
                else:
                    raise ValueError(f'Invalid beam section: {section}')
                #                  0        1       2        3        4        5       6        7       8
                props.append([prop.bf, prop.d, prop.tw, prop.tf, prop.ry, prop.A, prop.Ix, prop.My, prop.h])
            self.beam_properties[floor] = props

        # column properties
        self.column_properties = dict()
        for story, sections in self.columns.items():
            props = []
            for section in sections:
                if section.startswith(('W', 'w')):
                    prop = WSection(section, fy_beam)
                elif section.startswith(('HW', 'HM', 'HN')):
                    prop = GBSection(section, fy_beam)
                else:
                    raise ValueError(f'Invalid beam section: {section}')
                #                  0        1       2        3        4        5       6        7       8
                props.append([prop.bf, prop.d, prop.tw, prop.tf, prop.ry, prop.A, prop.Ix, prop.My, prop.h])
            self.column_properties[story] = props

        # RBS length
        self.RBS_length = dict()
        for floor in range(2, self.N + 2):
            if self.RBS_length_all:
                self.RBS_length[floor] = [self.RBS_length_all, self.RBS_length_all] * self.bays
            elif BC_connection in ['Full', 'Hinged']:
                self.RBS_length[floor] = [0, 0] * self.bays
            elif BC_connection == 'RBS':
                a, b, c = RBS_paras
                RBS_length_floor = []
                for id_bay in range(self.bays):  # Index for bay
                    bf_beam = self.beam_properties[floor][id_bay][0]
                    h_beam = self.beam_properties[floor][id_bay][1]
                    RBS_length_floor.append(a*bf_beam + 0.5*b*h_beam)
                    RBS_length_floor.append(a*bf_beam + 0.5*b*h_beam)
                    self.RBS_length[floor] = RBS_length_floor
            else:
                raise ValueError('[Error] 1')

    def _get_panel_zone_thickness(self):
        self.pz_thickness = dict()  # panel zone thickness (column web + doubler plate)
        for floor in range(2, self.N + 2):
            story = floor - 1
            pz_floor = []  # panel zone thickness of the floor
            for axis in range(1, self.axis + 1):
                id_axis = axis - 1
                if story not in self.column_splice:
                    tw = self.column_properties[story][id_axis][2]  # column web thickness
                else:
                    tw = self.column_properties[story+1][id_axis][2]
                db_plate = self.doubler_plate[floor][id_axis]  # doubler plate thickness
                pz_floor.append(tw + db_plate)
            self.pz_thickness[floor] = pz_floor





__version__ = '2.4'
print()
print('Tcl script generator for analysis of moment resisting frame using OpenSees')
print('Developed by: Wenchen Lie, Guangzhou University, China')
print(f'Version: {__version__}')
print('Data: 2024-03-16')




def check_int(value: int, min_max: list=None, pos=True, isNone=False, name=''):
    """check if a value is of the 'int' type and optionally within a given range"""
    name_ = f' `{name}`' if name else ''
    if not isNone and value is None:
        raise ValueError(f'Variable{name_} should not be of "None"')
    if isNone and value is None:
        return
    if not isinstance(value, int):
        raise ValueError(f'Variable{name_} should be of "int" type')
    if min_max:
        min_val, max_val = min_max
        if not min_val <= value <= max_val:
            raise ValueError(f'Variable{name_} should be within the range of [{min_val}, {max_val}]')
    if pos and value < 0:
        raise ValueError(f'The variable{name_} should be larger than 0')


def check_int_float(value: int | float, min_max: list=None, pos=True, isNone=False, name=''):
    """check if a value is of the 'int' or 'float' type and optionally within a given range"""
    name_ = f' `{name}`' if name else ''
    if not isNone and value is None:
        raise ValueError(f'Variable{name_} should not be of "None"')
    if isNone and value is None:
        return
    if not isinstance(value, (int, float)):
        raise ValueError(f'Variable{name_} should be of "int" or "float" type')
    if min_max:
        min_val, max_val = min_max
        if not min_val <= value <= max_val:
            raise ValueError(f'Variable{name_} should be within the range of [{min_val}, {max_val}]')
    if pos and value < 0:
        raise ValueError(f'The variable{name_} should be larger than 0')


def check_boolean(value: bool, isNone=False, name=''):
    """check if a value is of the 'boolean' type"""
    name_ = f' `{name}`' if name else ''
    if not isNone and value is None:
        raise ValueError(f'Variable{name_} should not be of "None"')
    if isNone and value is None:
        return
    if not isinstance(value, bool):
        raise ValueError(f'Variable{name_} should be of "bool" type')
    

def check_string(value: str, options: list=None, isNone=False, name=''):
    """check if a value is of the 'str' type"""
    name_ = f' `{name}`' if name else ''
    if not isNone and value is None:
        raise ValueError(f'Variable{name_} should not be of "None"')
    if isNone and value is None:
        return
    if not isinstance(value, str):
        raise ValueError(f'Variable{name_} should be of "str" type')
    if options and value not in options:
        raise ValueError(f'Variable{name_} should be within {options}')


def check_tuple(value: tuple, length=None, min_length=None, max_length=None, pos=True, isNone=False, name=''):
    """check if a value is of the 'tuple' type"""
    name_ = f' `{name}`' if name else ''
    if not isNone and value is None:
        raise ValueError(f'Variable{name_} should not be of "None"')
    if isNone and value is None:
        return
    if not isinstance(value, tuple):
        raise ValueError(f'The variable{name_} should be of "int" or "float" type')
    if length and len(value) != length:
        raise ValueError(f'The length of the tuple{name_} should be {length}')
    if min_length and len(value) < min_length:
        raise ValueError(f'The minimum length of the tuple{name_} should be {min_length}')
    if max_length and len(value) > max_length:
        raise ValueError(f'The maxmum length of the tuple{name_} should be {max_length}')
    if pos:
        for val in value:
            if val < 0:
                raise ValueError(f'The variable{name_} should be larger than 0')


def check_list(value: list, length=None, min_length=None, max_length=None, pos=True, isNone=False, name=''):
    """check if a value is of the 'list' type"""
    name_ = f' `{name}`' if name else ''
    if not isNone and value is None:
        raise ValueError(f'Variable{name_} should not be of "None"')
    if isNone and value is None:
        return
    if not isinstance(value, list):
        raise ValueError(f'The variable{name_} should be of "int" or "float" type')
    if length and len(value) != length:
        raise ValueError(f'The length of the list{name_} should be {length}')
    if min_length and len(value) < min_length:
        raise ValueError(f'The minimum length of the list{name_} should be {min_length}')
    if max_length and len(value) > max_length:
        raise ValueError(f'The maxmum length of the list{name_} should be {max_length}')
    if pos:
        for val in value:
            if val < 0:
                raise ValueError(f'The variable{name_} should be larger than 0')


def check_dict(value: dict, isNone=False, name=''):
    """check if a value is of the 'dict' type"""
    name_ = f' `{name}`' if name else ''
    if not isNone and value is None:
        raise ValueError(f'Variable{name_} should not be of "None"')
    if isNone and value is None:
        return
    if not isinstance(value, dict):
        raise ValueError(f'The variable{name_} should be of "int" or "float" type')



if __name__ == "__main__":
    a = None
    check_tuple(a, isNone=True, name='a')


import openseespy.opensees as ops
from typing import Dict
import matplotlib.pyplot as plt
from matplotlib.axes import Axes


def DisplayModel2D(model_data_init: tuple, ax: Axes):
    node_info_init: Dict[int, list[float, float]]
    ele_info_init: Dict[int, tuple[float, float, float, float]]
    model_size_init: tuple[float, float, float, float]
    node_info_init, ele_info_init, model_size_init = model_data_init
    ax.set_xlim(-1.4 * model_size_init[0], 1.4 * model_size_init[1])
    ax.set_ylim(-1.2 * model_size_init[2], 1.2 * model_size_init[3])
    for _, (x, y) in node_info_init.items():
        ax.scatter(x, y)
        plt.draw()
        # 无法在子线程中画图



def get_model_data():

    # 获取节点信息
    node_info: Dict[int, list[float, float]] = {}
    node_tags: list = ops.getNodeTags()
    for i, tag in enumerate(node_tags):
        node_info[tag] = ops.nodeCoord(tag)

    # 获取单元信息
    ele_info: Dict[int, tuple[float, float, float, float]] = {}
    ele_tags: list = ops.getEleTags()
    for i, tag in enumerate(ele_tags):
        inode, jnode = ops.eleNodes(tag)
        x1, y1 = node_info[inode]
        x2, y2 = node_info[jnode]
        ele_info[tag] = (x1, y1, x2, y2)

    # 获取模型尺寸范围
    x_max, x_min = 0, 0
    y_max, y_min = 0, 0
    for key, val in node_info.items():
        x, y = val
        x_max = max(x_max, x)
        x_min = min(x_min, x)
        y_max = max(y_max, y)
        y_min = min(y_min, y)
    model_size = (x_min, x_max, y_min, y_max)

    return node_info, ele_info, model_size




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .MRFhelper import Frame
from . import func
from typing import Literal


class ConnectionAndBoundary:

    def __init__(self, frame: Frame) -> None:
        """Step-4:
        Connection and boundary condition
        """
        self.frame = frame
        self.base_support = 'Fixed'
        self.beam_column_connection = 'Full'
        self.RBS_paras = (0.625, 0.75, 0.25)
        self.panel_zone_deformation = True
        self.soil_constraint = []
        self.rigid_disphragm = True


    def set_base_support(self, type: Literal['Fixed', 'Pinned']='Fixed'):
        """Defined column base support

        Args:
            type (str, optional):
            * "Fixed": Full constrained conncetion
            * "Pinned", Pinned connection
        """
        func.check_string(type, ['Fixed', 'Pinned'], 'type')
        self.base_support = type


    def set_beam_column_connection(self, type: Literal['Full', 'RBS', 'Hinged']='Full', a=0.625, b=0.75, c=0.25):
        """Define beam to column connection  

        Args:
            type (str, optional):
            * "Full" - Full restrained
            * "RBS" - Reduced beam section
            * "Hinged" - Hinged connection

            a, b, c: RBS parameters, default to 0.625, 0.75, and 0.25, respectively
        """
        func.check_int_float(a, [0, 10], name='a')
        func.check_int_float(b, [0, 10], name='b')
        func.check_int_float(c, [0, 1], name='c')
        func.check_string(type, ['Full', 'RBS', 'Hinged'], 'type')
        self.beam_column_connection = type
        self.RBS_paras = (a, b, c)


    def set_panel_zone_deformation(self, type: Literal[True, False]=True):
        """Define whether the panel zone deformation is consdiered

        Args:
            type (bool, optional): 
            * True - Use a parrallelogram model
            * False - Use a cruciform model
        """
        func.check_boolean(type, name='type')
        self.panel_zone_deformation = type

    def set_soil_constraint(self, floor: int):
        """Set soil constraint at specified floor

        Args:
            floor (int): floor number with soil constraint
        """
        func.check_int(floor, [1, self.frame.N + 1])
        if floor not in self.soil_constraint:
            self.soil_constraint.append(floor)
        

    def _finished(self):
        pass



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .MRFhelper import Frame
import numpy as np
from . import func


class LoadAndMaterial:
    g = 9800

    def __init__(self, frame: Frame) -> None:
        self.N = frame.N
        self.bays = frame.bays
        self.axis = frame.axis
        self.dead_load = dict()
        self.live_load = dict()
        self.cladding_load = dict()
        self.E = 206000  # Necessary
        self.fy_beam = None  # Necessary
        self.fy_column = None  # Necessary
        self.miu = 0.3  # Unnecessary
        self.cc_weight = {'Dead': 1.05, 'Live': 0.25, 'Cladding': 1.05}  # Combination coefficients of seismic weight, necessary
        self.cc_mass = {'Dead': 1, 'Live': 0, 'Cladding': 1}  # Combination coefficients of seismic mass, necessary
        for floor in range(2, frame.N + 2):
            self.dead_load[floor] = 1e-9
            self.live_load[floor] = 1e-9
            self.cladding_load[floor - 1] = 1e-9


    def set_dead_load(self, floor: list[int], load_per_area: list[int | float]):
        """Step-3:
        Set dead load for each floor

        Args:
            floor (list[int]): Numbers of floor
            load_per_area (list[int | float]): Values of dead load (surface load)
        """
        func.check_list(floor, length=self.N, name='floor')
        for val in floor:
            func.check_int(val, [2, self.N + 1], name='floor')
        func.check_list(load_per_area, length=self.N, name='load_per_area')
        for _, (a, b) in enumerate(zip(floor, load_per_area)):
            self.dead_load[a] = b


    def set_live_load(self, floor: list[int], load_per_area: list[int | float]):
        """Set live load for each floor

        Args:
            floor (list[int]): Numbers of floor
            load_per_area (list[int | float]): Values of live load (surface load)
        """
        func.check_list(floor, length=self.N, name='floor')
        for val in floor:
            func.check_int(val, [2, self.N + 1], name='floor')
        func.check_list(load_per_area, length=self.N, name='load')
        for _, (a, b) in enumerate(zip(floor, load_per_area)):
            self.live_load[a] = b


    def set_cladding_load(self, story: list[int], load_per_area: list[int | float]):
        """Set cladding load for each story

        Args:
            story (list[int]): Numbers of story
            load_per_area (list[int | float]): Values of cladding load (surface load)
        """
        func.check_list(story, length=self.N, name='story')
        for val in story:
            func.check_int(val, [1, self.N], name='story')
        func.check_list(load_per_area, length=self.N, name='load')
        for _, (a, b) in enumerate(zip(story, load_per_area)):
            self.cladding_load[a] = b


    def set_material(self, E: int | float, fy_beam: int | float, fy_column: int | float, miu: float=0.3):
        """Set material properties for all steel components

        Args:
            E (int | float): Youngs modulus
            fy_beam (int | float): Nominal yield strength of beams
            fy_column (int | float): Nominal yield strength of column
            miu (float, optional): possion ratio
        """
        func.check_int_float(E, name='E')
        func.check_int_float(fy_beam, name='fy_beam')
        func.check_int_float(fy_column, name='fy_column')
        func.check_int_float(miu, [0, 0.5], name='miu')
        self.E = E
        self.fy_beam = fy_beam
        self.fy_column = fy_column
        self.miu = miu


    def set_weight_combination_coefficients(self, coefficients: dict):
        """Set seicmic weight combination coefficients  
        Note: Load type should be within ['Dead', 'Live', 'Cladding']

        Args:
            coefficients (dict): {load type: combination coefficient}
        """
        func.check_dict(coefficients, name='coefficients')
        for key, val in coefficients.items():
            if not key in ['Dead', 'Live', 'Cladding']:
                raise ValueError("Load type should be within ['Dead', 'Live', 'Cladding']")
            func.check_int_float(val, name='coefficient')
            self.cc_weight[key] = val


    def set_mass_combination_coefficients(self, coefficients: dict):
        """Set seicmic mass combination coefficients  
        Note: Load type should be within ['Dead', 'Live', 'Cladding']

        Args:
            coefficients (dict): {load type: combination coefficient}
        """
        func.check_dict(coefficients, name='coefficients')
        for key, val in coefficients.items():
            if not key in ['Dead', 'Live', 'Cladding']:
                raise ValueError("Load type should be within ['Dead', 'Live', 'Cladding']")
            func.check_int_float(val, name='coefficient')
            self.cc_mass[key] = val


    def _finished(self):
        if self.E is None:
            raise ValueError("Young's modulus has not been difined")
        if self.fy_beam is None:
            raise ValueError("Nominal yield strength of beams has not been difined")
        if self.fy_column is None:
            raise ValueError("Nominal yield strength of columns has not been difined")
        if not self.cc_weight:
            raise ValueError("Combination coefficients of seismic weight have not been difined")
        if not self.cc_mass:
            raise ValueError("Combination coefficients of seismic mass have not been difined")
        floor_set = set([2 + i for i in range(self.N)])
        floor_defined_dead = set(self.dead_load.keys())
        floor_defined_live = set(self.live_load.keys())
        if floor_set == floor_defined_dead:
            return
        if floor_set - floor_defined_dead:
            raise ValueError(f'Dead load for floor {floor_set - floor_defined_dead} have not been defined')
        if floor_defined_dead - floor_set:
            raise ValueError(f'Floor {floor_defined_dead - floor_set} are not existed')
        if floor_set - floor_defined_live:
            raise ValueError(f'Dead load for floor {floor_set - floor_defined_live} have not been defined')
        if floor_defined_live - floor_set:
            raise ValueError(f'Floor {floor_defined_live - floor_set} are not existed')
        self._calculate_load()
        

    def _calculate_load(self, frame: Frame):
        """Calculate load and mass
        * F_node (dict): Beam-column joint force ({floor: force})
        * F_grav (dect): Leaning column joint force ({floor: force})
        * mass_node (dict): Beam-column joint mass ({floor: mass})
        * mass_grav (dect): Leaning column joint mass ({floor: mass})
        """
        plane_dimensions = frame.BuildingGeometry.plane_dimensions
        N_MF = frame.BuildingGeometry.MF_number
        area_ex = frame.BuildingGeometry.exterior_column_tributary_area
        area_in = frame.BuildingGeometry.interior_column_tributary_area
        h_list = frame.BuildingGeometry.story_height
        # dead load
        DL_node = dict()
        DL_grav = dict()
        S_3D = plane_dimensions[0] * plane_dimensions[1]  # area of 3D building plane
        for floor, DL_value in self.dead_load.items():
            DL_3D = S_3D * DL_value  # Dead load (N) for each floor of the 3D building
            DL = DL_3D / N_MF  # Dead load (N) for each floor of the 2D considered frame
            r_rest = 1
            DL_floor = []
            for id_axis in range(self.axis):
                if id_axis == 0 or id_axis == self.axis - 1:
                    r = area_ex[0] * area_ex[1] / (S_3D / N_MF)  # area ratio for external columns
                else:
                    r = area_in[0] * area_in[1] / (S_3D / N_MF)  # area ratio for internal columns
                r_rest -= r
                DL_floor.append(DL * r)
            DL_grav[floor] = DL * r_rest
            DL_node[floor] = DL_floor
        # live load (same as the dead load)
        LL_node = dict()
        LL_grav = dict()
        S_3D = plane_dimensions[0] * plane_dimensions[1]  # area of 3D building plane
        for floor, LL_value in self.live_load.items():
            LL_3D = S_3D * LL_value  # Dead load (N) for each floor of the 3D building
            LL = LL_3D / N_MF  # Dead load (N) for each floor of the 2D considered frame
            r_rest = 1
            LL_floor = []
            for id_axis in range(self.axis):
                if id_axis == 0 or id_axis == self.axis - 1:
                    r = area_ex[0] * area_ex[1] / (S_3D / N_MF)  # area ratio for external columns
                else:
                    r = area_in[0] * area_in[1] / (S_3D / N_MF)  # area ratio for internal columns
                r_rest -= r
                LL_floor.append(LL * r)
            LL_grav[floor] = LL * r_rest
            LL_node[floor] = LL_floor
        # cladding load
        CL_node = dict()
        CL_grav = dict()
        for story, CL_value in self.cladding_load.items():
            floor = story + 1
            story_b, story_t = floor - 1, floor  # story number
            CL_floor = []
            h = h_list[story_b-1]/2 if floor==self.N+1 else (h_list[story_b-1] + h_list[story_t-1])/2  # tributary heigth of each floor
            S_rest = plane_dimensions[0] * h
            for id_axis in range(self.axis):
                if id_axis == 0 or id_axis == self.axis - 1:  # if external node
                    if floor == self.N + 1:  # if top floor
                        S_node = area_ex[0] * h_list[story_b-1] / 2  # Node tributary facade area
                    else:
                        S_node = area_ex[0] * (h_list[story_b-1] + h_list[story_t-1]) / 2  # Node tributary facade area
                else:
                    if floor == self.N + 1:  # if top floor
                        S_node = area_in[0] * h_list[story_b-1] / 2  # Node tributary facade area
                    else:
                        S_node = area_in[0] * (h_list[story_b-1] + h_list[story_t-1]) / 2  # Node tributary facade area
                S_rest -= S_node
                CL_floor.append(S_node * CL_value)
            CL_node[floor] = CL_floor
            CL_grav[floor] = S_rest * CL_value + plane_dimensions[1] * h * CL_value  # Consider the Y-direction cladding load
            # CL_grav[floor] = S_rest * CL_value   # TODO Did not consider the Y-direction cladding load
        # node force
        F_node = dict()
        F_grav = dict()
        mass_node = dict()
        mass_grav = dict()
        for floor in range(2, 2 + self.N):
            DL_floor = np.array(DL_node[floor])
            LL_floor = np.array(LL_node[floor])
            CL_floor = np.array(CL_node[floor])
            F_floor = DL_floor * self.cc_weight['Dead'] + LL_floor * self.cc_weight['Live'] + CL_floor + self.cc_weight['Cladding']
            mass_floor = (DL_floor * self.cc_mass['Dead'] + LL_floor * self.cc_mass['Live'] + CL_floor + self.cc_mass['Cladding']) / self.g
            F_floor = list(F_floor)
            mass_floor = list(mass_floor)
            F_node[floor] = F_floor
            mass_node[floor] = mass_floor
            F_grav[floor] = DL_grav[floor] * self.cc_weight['Dead'] + LL_grav[floor] * self.cc_weight['Live'] + CL_grav[floor] * self.cc_weight['Cladding']
            mass_grav[floor] = (DL_grav[floor] * self.cc_mass['Dead'] + LL_grav[floor] * self.cc_mass['Live'] + CL_grav[floor] * self.cc_mass['Cladding']) / self.g
        # TODO Beam and column component masses have not been considered, but FM2D also did not consider.
        self.F_node, self.F_grav, self.mass_node, self.mass_grav = F_node, F_grav, mass_node, mass_grav
            

    def _calculate_PPy(self, frame: Frame, PPy_scale: float=1.25):
        """Calculate the axial compression ratio of columns
        * PPy (dict): Compression ratio ({story: ratio})
        * PPy_scale (float): Scale factor of compression ratio to consider the overturning effect,
        defaults to 1.25

        Args:
            PPy_scale (float, optional): Scale factor of compression ratio, defaults to 1.25.
        """
        props_col = frame.StructuralComponents.column_properties
        P_col = dict()  # Axial forces of columns
        P_temp = np.zeros(self.axis)
        for story in list(1+i for i in range(self.N))[::-1]:
            floor = story + 1
            P_temp += np.array(self.F_node[floor])
            P_col[story] = P_temp.tolist()
        PPy = dict()  # Column axial compression ratio
        for story in range(1, self.N + 1):
            prop_floor = props_col[story]  # column properties for each story
            PPy_story_b, PPy_story_t = [], []
            for axis in range(1, self.axis + 1):
                id_axis = axis - 1
                A = prop_floor[id_axis][5]  # Cross section aera of column
                P = P_col[story][id_axis]  # Axial force of column
                PPy_story_b.append(P / (A * self.fy_column))  # Axial compression ratio of columns on the story
                if story not in frame.StructuralComponents.column_splice:
                    PPy_story_t.append(P / (A * self.fy_column))  # Axial compression ratio of columns on the story
                else:
                    A_t = props_col[story+1][id_axis][5]  # Cross section aera of column
                    PPy_story_t.append(P / (A_t * self.fy_column))  # Axial compression ratio of columns on the story
            PPy[f'{story}b'] = PPy_story_b
            PPy[f'{story}t'] = PPy_story_t
        self.PPy, self.PPy_scale = PPy, PPy_scale

                




import numpy as np
import openseespy.opensees as ops
import time
import matplotlib.pyplot as plt
# from subroutines import DisplayModel2D


def TimeHistorySolver(
        dt_init: float, duration: float, story_heights: list,
        ctrl_nodes: list,  CollapseDrift: float, MaxAnalysisDrift: float,
        GMname: str, maxRunTime: float, ShowAnimation: bool,
        min_factor: float=1e-6, max_factor: float=1) -> int:
    """This solver is used to perform time history analysis for frame structure.

    Args:
        dt_init (float): Ground motion step
        duration (float): Ground motion duration
        story_heights (list): Heights of story
        ctrl_nodes (list): Controlled nodes of story
        CollapseDrift (float): Drift ratio that sign the frame as collapse
        MaxAnalysisDrift (float): Drift ratio that analysis termination
        GMname (str): Ground motion name
        maxRunTime (float): Maximum run time (second)
        ShowAnimation (bool): Whether to display the building deformation
        print_result (bool): Whether to print analysis information
        min_factor (float): Factor to control the adaptive time step
        max_factor (float): Factor to control the adaptive time step
    
    Return:
        int: 1 - Analysis finished, the structure did not collapse,
        2 - The structure collapsed,
        3 - Cannot converge,
        4 - Exceeding maximum running time
    """

    algorithms = [("KrylovNewton",), ("NewtonLineSearch",), ("Newton",), ("SecantNewton",)]
    algorithm_id = 0
    ops.wipeAnalysis()
    ops.constraints("Plain")
    ops.numberer("RCM")
    ops.system("UmfPack")
    ops.test("EnergyIncr", 1.0e-3, 100)
    ops.algorithm("KrylovNewton")
    ops.integrator("Newmark", 0.5, 0.25)
    ops.analysis("Transient")

    collapse_flag = False
    print_collapse = True
    factor = 1
    start_time = time.time()
    nstep = 0
    dt = dt_init
    SDRs = np.zeros((10, len(story_heights)))
    SDR_roof = [0] * 10
    collapseTime = 60
    collapseStart = 0
    while True:
        if time.time() - start_time > maxRunTime:
            print("Exceeding maximum running time")
            return 3, ops.getTime(), collapse_flag, SDRs, SDR_roof
        if collapse_flag and time.time() - collapseStart > collapseTime:
            return 2, ops.getTime(), collapse_flag, SDRs, SDR_roof
        ok = ops.analyze(1, dt)
        if ok == 0:
            collapse_flag, maxAna_flag, SDRs_i, SDR_roof_i = SDR_tester(
                story_heights, ctrl_nodes, CollapseDrift, MaxAnalysisDrift, GMname)
            if collapse_flag and collapseStart == 0:
                collapseStart = time.time()
            try:
                SDRs = np.vstack((SDRs, SDRs_i))
            except:
                print(SDRs)
                print(SDRs_i)
                raise ValueError()
            SDR_roof.append(SDR_roof_i)
            if ops.getTime() >= duration:
                print("Analysis finished")
                # if ShowAnimation:
                #     plt.ioff()
                return 1, ops.getTime(), collapse_flag, SDRs, SDR_roof
            if maxAna_flag:
                print("Analysis finished, the structure collapsed")
                # if ShowAnimation:
                #     plt.ioff()
                return 1, ops.getTime(), collapse_flag, SDRs, SDR_roof
            if collapse_flag and print_collapse:
                print("The structure was collapse")
                print_collapse = False
            factor_old = factor
            factor *= 2
            factor = min(factor, max_factor)
            if factor_old < factor:
                print(f"---- Enlarged factor: {factor}, Time: {ops.getTime()}")
            algorithm_id -= 1
            algorithm_id = max(0, algorithm_id)
            # algorithm_id += 1
            # algorithm_id == 4
            # algorithm_id = 0
        else:
            factor *= 0.5
            if factor < min_factor:
                factor = min_factor
                algorithm_id += 1
                if algorithm_id == 4:
                    print("Cannot converge")
                    return 2, ops.getTime(), collapse_flag, SDRs, SDR_roof
                print(f"-------- Switched algorithm:", *algorithms[algorithm_id], f'Time: {ops.getTime()}')
            print(f"---- Reduced factor: {factor}, Time: {ops.getTime()}")
        dt = dt_init * factor
        if dt + ops.getTime() > duration:
            dt = duration - ops.getTime()
        ops.algorithm(*algorithms[algorithm_id])
        nstep += 1
        # if nstep == 20:
        #     break
            

def SDR_tester(story_heights: list, ctrl_nodes: list,
               CollapseDrift: float, MaxAnalysisDrift: float,
               GMname: str) -> tuple[bool, bool, list, float]:
    """
    return (tuple[bool, bool]): Exceeding CollapseDrift? Exceeding MaxAnalysisDrift? 
    """
    if CollapseDrift > MaxAnalysisDrift:
        raise ValueError('`MaxAnalysisDrift` should be larger than `CollapseDrift`')

    SDRs = []
    SDR_roof = ops.nodeDisp(ctrl_nodes[len(story_heights) - 1])[0] / sum(story_heights)
    collapse = (False, False)
    for i, h in enumerate(story_heights):
        if i == 0:
            disp_b = 0
            disp_t = ops.nodeDisp(ctrl_nodes[0])[0]
        else:
            disp_b = ops.nodeDisp(ctrl_nodes[i - 1])[0]
            disp_t = ops.nodeDisp(ctrl_nodes[i])[0]
        SDR = (disp_t - disp_b) / h
        SDRs.append(SDR)
        if SDR >= MaxAnalysisDrift:
            collapse = (True, True)  # TODO
        if SDR >= CollapseDrift:
            collapse = (True, False)  # TODO 这两个if应调换
    return *(collapse), SDRs, SDR_roof




if __name__ == "__main__":
    t = ops.getTime()













# -------------------------- Construct beam hinge model --------------------------
#
# Args (14):
# ---------------
# SpringID       Zero length element ID
# NodeI          Node i ID
# NodeJ          Node j ID
# E              Young's modulus
# Ix             Moment of inertia of section
# d              Section depth
# htw            Web slenderness ratio
# ry             Radius of gyration
# L              Member Length
# Lb             Unbraced length
# My             Effective Yield Moment
# PPy            Axial load ratio due to gravity
# SF_PPy         Scale factor of axial load ratio (due to overturning effect)
# type_          Column base pinned connection (1: fixed, 2: pinned, 3: elastic)
# check          Print IMK model parameters
# 
# ---------------
# Reference:
# [1] Proposed Updates to the ASCE 41 Nonlinear Modeling Parameters for Wide-Flange Steel Columns in Support of Performance-Based Seismic Engineering
# ---------------
# Written by: Wenchen Lie, Guangzhou University, China
# Date: 2024-03-26
# --------------------------------------------------------------------------------

import openseespy.opensees as ops
from typing import Literal


def ColumnHinge(
    SpringID: int, NodeI: int, NodeJ: int,
    E: float, Ix: float, d: float, htw: float, ry: float,
    L: float, Lb: float, My: float, PPy: float,
    SF_PPy: float, type_: Literal[1, 2, 3], check: bool=None):
    
    n = 10.0
    c1 = 1.0
    c2 = 1.0
    PPy *= float(SF_PPy)  # Enlarge PPy for considering overturning effect

    K = (n + 1.0) * 6 * E * Ix / L

    theta_p = 294.0 * ((htw) ** -1.700) * ((Lb / ry) ** -0.700) * ((1 - PPy) ** 1.600)  # Eq. (7)
    theta_pc = 90.0 * ((htw) ** -0.800) * ((Lb / ry) ** -0.800) * ((1 - PPy) ** 2.500)  # Eq. (9)
    if theta_p > 0.20:
        theta_p = 0.2
    if theta_pc > 0.30:
        theta_pc = 0.3

    if PPy <= 0.35:
        Lamda = 25500.0 * ((htw) ** -2.140) * ((Lb / ry) ** -0.530) * ((1 - PPy) ** 4.920)
    else:
        Lamda = 268000.0 * ((htw) ** -2.300) * ((Lb / ry) ** -1.300) * ((1 - PPy) ** 1.190)
    if Lamda < 3.0:
        Lamda = 3.0

    if PPy < 0.2:
        My = (1.15 / 1.1) * My * (1 - PPy / 2)
    else:
        My = (1.15 / 1.1) * My * (9 / 8) * (1 - PPy)
    # Eq. (2)

    McMy = 12.5 * ((htw) ** -0.200) * ((Lb / ry) ** -0.400) * ((1 - PPy) ** 0.400)  # Eq. (3)
    if McMy < 1.0:
        McMy = 1.0
    if McMy > 1.3:
        McMy = 1.3

    theta_u = 0.15
    D = 1.0
    Res = 0.5 - 0.4 * PPy  # Eq. (5)
    c = 1.0

    if type_ == 1:
        ops.uniaxialMaterial("IMKBilin", SpringID, K, theta_p, theta_pc, theta_u, My, McMy, Res, theta_p, theta_pc, theta_u, My, McMy, Res, Lamda, 0.9 * Lamda, 0.9 * Lamda, c, c, c, D, D)
        ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", 99, 99, SpringID, "-dir", 1, 2, 6)
        # ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", SpringID, "-dir", 6)
        # ops.equalDOF(NodeI, NodeJ, 1, 2)
    elif type_ == 2:
        ops.uniaxialMaterial("IMKBilin", SpringID, K, theta_p, theta_pc, theta_u, My, McMy, Res, theta_p, theta_pc, theta_u, My, McMy, Res, Lamda, 0.9 * Lamda, 0.9 * Lamda, c, c, c, D, D)
        ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", 99, 99, 9, "-dir", 1, 2, 6)
        # ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", 9, "-dir", 6)
        # ops.equalDOF(NodeI, NodeJ, 1, 2)
    elif type_ == 3:
        # Elastic
        ops.uniaxialMaterial("Elastic", SpringID, K)
        ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", 99, 99, SpringID, "-dir", 1, 2, 6)
    if check:
        print(f"{check}:\nKs: {K}, My: {My}, theta_p: {theta_p}, theta_pc: {theta_pc}, Res: {Res}")





# -------------------------- Construct panel zone model --------------------------
#
# Args (16):
# ---------------
# Floor          FLoor number
# Axis           Axis number
# X, Y           Coordiate of panel zone center
# E              Young's modulus
# mu             Possion ration
# fy             Yield strength
# A_stiff        Cross section area of rigid link
# I_stiff        Moment of inertia of rigid link
# d_col          Column section depth
# d_beam         Beam section depth
# tp             Panel zone thickness (column web + doubler plate)
# tf             Column flange thickness
# bf             Column flange width
# transfTag      Geometric transformation ID
# type_          Modeling approach (1: Parallelogram, 2: Cruciform)
# position       Panel zone position
#                   I: Internal
#                   L: Left
#                   R: Right
#                   T: top
#                   LT: Left top
#                   RT: Right top
# check          Print Hysteretic model parameters
# 
# ---------------
# Written by: Wenchen Lie, Guangzhou University, China
# Date: 2024-03-26
# --------------------------------------------------------------------------------

import openseespy.opensees as ops
from typing import Literal


def PanelZone(
    Floor: int, Axis: int,
    X: float, Y: float,
    E: float, mu: float, fy: float,
    A_stiff: float, I_stiff: float,
    d_col: float, d_beam: float, tp: float, tf: float,
    bf: float, transfTag: Literal[1, 2, 3],
    type_: Literal[1, 2],
    position: Literal['L', 'T', 'LT', 'RT'], check: bool=None):
    # node ID
    node_C = 11000000 + Floor * 10000 + Axis * 100  # 11FFAA01
    node_B = node_C + 1  # 11FFAA01
    node_L = node_C + 2  # 11FFAA02
    node_T = node_C + 3  # 11FFAA03
    node_R = node_C + 4  # 11FFAA04
    node_BL = node_C + 5  # 11FFAA05
    node_LB = node_C + 6  # 11FFAA06
    node_LT = node_C + 7  # 11FFAA07
    node_TL = node_C + 8  # 11FFAA08
    node_TR = node_C + 9  # 11FFAA09
    node_RT = node_C + 10  # 11FFAA10
    node_RB = node_C + 11  # 11FFAA11
    node_BR = node_C + 12  # 11FFAA12

    # Construct nodes
    if type_ == 1:
        ops.node(node_B, X, Y - 0.5 * d_beam)
        ops.node(node_R, X + 0.5 * d_col, Y)
        ops.node(node_T, X, Y + 0.5 * d_beam)
        ops.node(node_L, X - 0.5 * d_col, Y)
    elif type_ == 2:
        ops.node(node_B, X, Y - 0.5 * d_beam)
        ops.node(node_R, X + 0.5 * d_col, Y)
        if type_ != "L" and type_ != "LT":
            ops.node(node_L, X - 0.5 * d_col, Y)
        if type_ != "T" and type_ != "LT" and type_ != "RT":
            ops.node(node_T, X, Y + 0.5 * d_beam)
    if type_ == 1:
        ops.node(node_BL, X - 0.5 * d_col, Y - 0.5 * d_beam)
        ops.node(node_LB, X - 0.5 * d_col, Y - 0.5 * d_beam)
        ops.node(node_LT, X - 0.5 * d_col, Y + 0.5 * d_beam)
        ops.node(node_TL, X - 0.5 * d_col, Y + 0.5 * d_beam)
        ops.node(node_TR, X + 0.5 * d_col, Y + 0.5 * d_beam)
        ops.node(node_RT, X + 0.5 * d_col, Y + 0.5 * d_beam)
        ops.node(node_RB, X + 0.5 * d_col, Y - 0.5 * d_beam)
        ops.node(node_BR, X + 0.5 * d_col, Y - 0.5 * d_beam)
    elif type_ == 2:
        ops.node(node_C, X, Y)

    # Rigid elements ID
    ele_B = 11000000 + Floor * 10000 + Axis * 100 + 1
    ele_L = ele_B + 1
    ele_T = ele_B + 2
    ele_R = ele_B + 3
    ele_BL = 11000000 + Floor * 10000 + Axis * 100 + 1
    ele_LB = ele_BL + 1
    ele_LT = ele_BL + 2
    ele_TL = ele_BL + 3
    ele_TR = ele_BL + 4
    ele_RT = ele_BL + 5
    ele_RB = ele_BL + 6
    ele_BR = ele_BL + 7

    # Construct rigid elements
    if type_ == 1:
        ops.element("elasticBeamColumn", ele_BL, node_B, node_BL, A_stiff, E, I_stiff, transfTag)
        ops.element("elasticBeamColumn", ele_LB, node_LB, node_L, A_stiff, E, I_stiff, transfTag)
        ops.element("elasticBeamColumn", ele_LT, node_L, node_LT, A_stiff, E, I_stiff, transfTag)
        ops.element("elasticBeamColumn", ele_TL, node_TL, node_T, A_stiff, E, I_stiff, transfTag)
        ops.element("elasticBeamColumn", ele_TR, node_T, node_TR, A_stiff, E, I_stiff, transfTag)
        ops.element("elasticBeamColumn", ele_RT, node_RT, node_R, A_stiff, E, I_stiff, transfTag)
        ops.element("elasticBeamColumn", ele_RB, node_R, node_RB, A_stiff, E, I_stiff, transfTag)
        ops.element("elasticBeamColumn", ele_BR, node_BR, node_B, A_stiff, E, I_stiff, transfTag)
    elif type_ == 2:
        ops.element("elasticBeamColumn", ele_B, node_C, node_B, A_stiff, E, I_stiff, transfTag)
        ops.element("elasticBeamColumn", ele_R, node_C, node_R, A_stiff, E, I_stiff, transfTag)
        if position != "L" and position != "LT":
            ops.element("elasticBeamColumn", ele_L, node_C, node_L, A_stiff, E, I_stiff, transfTag)
        if position != "T" and position != "LT" and position != "RT":
            ops.element("elasticBeamColumn", ele_T, node_C, node_T, A_stiff, E, I_stiff, transfTag)

    # Restrain DOF at panel corners
    if type_ == 1:
        ops.equalDOF(node_RB, node_BR, 1, 2)
        ops.equalDOF(node_BL, node_LB, 1, 2)
        ops.equalDOF(node_LT, node_TL, 1, 2)
        ops.equalDOF(node_TR, node_RT, 1, 2)

    # Calculating shear behavior
    G = E / (2 * (1 + mu))  # shear modulus
    # 1
    V1 = 0.55 * fy * d_col * tp  # yield force
    M1 = V1 * d_beam  # yield moment
    gamma1 = fy / (G * pow(3, 0.5))  # yield rotation
    K1 = 0.95 * d_col * tp * G  # elastic stiffness
    Kgamma1 = M1 / gamma1  # elastic rotational stiffness
    # 2
    gamma2 = 4.0 * gamma1
    V2 = 0.55 * fy * d_col * tp * (1 + (3.0 * bf * tf ** 2) / (d_beam * d_col * tp))
    M2 = V2 * d_beam
    # 3
    Kgamma3 = 0.03 * Kgamma1
    gamma3 = 100. * gamma1
    M3 = M2 + (gamma3 - gamma2) * Kgamma3
    # Ultimate
    gammaU = 0.3

    # Construct panel zone spring
    if type_ == 1:
        spring_id = 11000000 + Floor * 10000 + Axis * 100
        ops.uniaxialMaterial("Hysteretic", spring_id + 1, M1, gamma1, M2, gamma2, M3, gamma3, -M1, -gamma1, -M2, -gamma2,
                        -M3, -gamma3, 0.25, 0.75, 0., 0., 0.)
        ops.uniaxialMaterial("MinMax", spring_id, spring_id + 1, "-min", -gammaU, "-max", gammaU)
        ops.element("zeroLength", spring_id, node_TR, node_RT, "-mat", spring_id, "-dir", 6)

    if check:
        print(check)
        print(f"M1: {M1}, M2: {M2}, M3: {M3}")
        print(f"gamma1: {gamma1}, gamma2: {gamma2}, gamma3: {gamma3}")



import json
from pathlib import Path
from .BuildingGeometry import BuildingGeometry
from .StructuralComponents import StructuralComponents
from .ConnectionAndBoundary import ConnectionAndBoundary
from .LoadAndMaterial import LoadAndMaterial
from .UserCommand import UserCommand
from . import WriteInfo
from . import WriteScript
from . import __version__


VERSION = __version__

class Frame:
    version = VERSION

    def __init__(self, frame_name: str, notes: str=None):
        """Use this class to define structural parameters of steel moment resisting frame (MRF)
        and write tcl script for analysis using OpenSees. Make sure all the units are [N, mm, t].

        Args:
            frame_name (str): Give a name for the considered frame
            notes (str, optional): Optional notes
        """
        if not frame_name.isidentifier():
            raise ValueError(f'Illegal frame name: {frame_name}')
        self.frame_name = frame_name
        self.notes = notes
        self.BuildingGeometry = BuildingGeometry()

    def step1_finished(self):
        """Complete define building geometry"""
        self.BuildingGeometry._finish()
        self.N = self.BuildingGeometry.N  # number of stories
        self.bays = self.BuildingGeometry.bays  # number of bays
        self.axis = self.BuildingGeometry.axis  # number of axes
        self.StructuralComponents = StructuralComponents(self)

    def step2_finished(self):
        """Complete define structural components"""
        self.StructuralComponents._finish()
        self.LoadAndMaterial = LoadAndMaterial(self)

    def step3_finished(self):
        """Complete define load and material"""
        self.LoadAndMaterial._finished()
        self.ConnectionAndBoundary = ConnectionAndBoundary(self)

    def step4_finished(self):
        """Complete define connection and boundary"""
        self.ConnectionAndBoundary._finished()
        self.recorders = {
            'Reactions': True,
            'Drift': True,
            'FloorAccel': True,
            'FloorVel': True,
            'FloorDisp': True,
            'ColumnForce': True,
            'ColumnHinge': True,
            'BeamHinge': True,
            'PanelZone': True
        }
        self.UserComment = UserCommand()

    def all_steps_finished(self):
        """Calculate simulation parameters and write them into tcl script"""
        self.StructuralComponents._get_section_properties(self)
        self.StructuralComponents._get_panel_zone_thickness()
        self.LoadAndMaterial._calculate_load(self)
        self.LoadAndMaterial._calculate_PPy(self)
        self.dict_info = WriteInfo.write_info_to_dict(self)

    def generate_tcl_script(self, dir_):
        """Write tcl script"""
        self.output_path = Path(dir_)
        if not self.output_path.exists():
            Path.mkdir(self.output_path)
        self.builiding_info = WriteInfo.write_info_to_tcl(self)
        WriteScript.WriteScript(self)


def from_json(file: str | Path) -> Frame:
    """Get an available model from json file

    Args:
        file (str | Path): file path

    Returns:
        Frame: Object of `Frame`
    """
    file = Path(file)
    if not file.exists():
        raise FileExistsError(f'file not found')
    with open(file, 'r') as f:
        dict_info = json.load(f)
    frame = Frame(dict_info['name'])
    # Step 1
    frame.BuildingGeometry.story_height = dict_info['BuildingGeometry']['story_height']
    frame.BuildingGeometry.bay_length = dict_info['BuildingGeometry']['bay_length']
    frame.BuildingGeometry.plane_dimensions = tuple(dict_info['BuildingGeometry']['plane_dimensions'])
    frame.BuildingGeometry.MF_number = dict_info['BuildingGeometry']['MF_number']
    frame.BuildingGeometry.exterior_column_tributary_area = tuple(dict_info['BuildingGeometry']['exterior_column_tributary_area'])
    frame.BuildingGeometry.interior_column_tributary_area = tuple(dict_info['BuildingGeometry']['interior_column_tributary_area'])
    frame.step1_finished()
    # Step 2
    for FF, sections in dict_info['StructuralComponents']['beams'].items():
        frame.StructuralComponents.set_beams(int(FF), sections)
    for SS, sections in dict_info['StructuralComponents']['columns'].items():
        frame.StructuralComponents.set_columns(int(SS), sections)
    for FF, thickness in dict_info['StructuralComponents']['set_doubler_plate'].items():
        frame.StructuralComponents.set_doubler_plate(int(FF), thickness)
    frame.StructuralComponents.set_column_splice(*dict_info['StructuralComponents']['column_splice'])
    frame.StructuralComponents.set_beam_splice(*dict_info['StructuralComponents']['beam_splice'])
    if dict_info['StructuralComponents']['RBS_length']:
        frame.StructuralComponents.set_RBS_length(dict_info['StructuralComponents']['RBS_length'])
    frame.step2_finished()
    # Step 3
    FF_ls = list(int(i) for i in dict_info['LoadAndMaterial']['dead_load'].keys())
    SS_ls = list(int(i) for i in dict_info['LoadAndMaterial']['clading_load'].keys())
    frame.LoadAndMaterial.set_dead_load(FF_ls, list(dict_info['LoadAndMaterial']['dead_load'].values()))
    frame.LoadAndMaterial.set_live_load(FF_ls, list(dict_info['LoadAndMaterial']['live_load'].values()))
    frame.LoadAndMaterial.set_cladding_load(SS_ls, list(dict_info['LoadAndMaterial']['clading_load'].values()))
    frame.LoadAndMaterial.set_weight_combination_coefficients(dict_info['LoadAndMaterial']['weight_combination_coefficients'])
    frame.LoadAndMaterial.set_mass_combination_coefficients(dict_info['LoadAndMaterial']['mass_combination_coefficients'])
    material = dict_info['LoadAndMaterial']['material']
    frame.LoadAndMaterial.set_material(material['E'], material["fy_beam"], material["fy_column"], material["miu"])
    frame.step3_finished()
    # Step 4
    frame.ConnectionAndBoundary.set_base_support(dict_info['ConnectionAndBoundary']['base_support'])
    frame.ConnectionAndBoundary.set_beam_column_connection(dict_info['ConnectionAndBoundary']['beam_column_connection'])
    frame.ConnectionAndBoundary.set_panel_zone_deformation(dict_info['ConnectionAndBoundary']['panel_zone_deformation'])
    if dict_info['ConnectionAndBoundary']['soil_constraint']:
        frame.ConnectionAndBoundary.set_soil_constraint(dict_info['ConnectionAndBoundary']['soil_constraint'])
    frame.step4_finished()
    # finish
    frame.all_steps_finished()
    frame.dict_info['References'] = dict_info['References']
    return frame


# -------------------------- Construct beam hinge model --------------------------
#
# Args (17):
# ---------------
# SpringID       Zero length element ID
# NodeI          Node i ID
# NodeJ          Node j ID
# E              Young's modulus
# fy             Yield strength
# Ix             Moment of inertia of section
# d              Section depth
# htw            Web slenderness ratio
# bftf           Flange slenderness ratio
# ry             Radius of gyration
# L              Member Length
# Ls             Shear Span
# Lb             Unbraced length
# My             Effective Yield Moment
# type_          Type of beam-to-column connection
#                   1: Reduced beam section (RBS)
#                   2: Other than RBS
#                   3: Pinned
#                   4: Elastic
# check          If given, print IMK model parameters
# 
# ---------------
# Reference:
# [1] Deterioration Modeling of Steel Components in Support of Collapse Prediction of Steel Moment Frames under Earthquake Loading
# ---------------
# Written by: Wenchen Lie, Guangzhou University, China
# Date: 2024-03-26
# --------------------------------------------------------------------------------
import openseespy.opensees as ops
from typing import Literal


def BeamHinge(
    SpringID: int, NodeI: int, NodeJ: int,
    E: float, fy: float,
    Ix: float, d: float, htw: float, bftf: float,
    ry: float, L: float, Ls: float, Lb: float,
    My: float, type_: Literal[1, 2, 3], check: bool=None):
    
    n = 10.0
    c1 = 1.0
    c2 = 1.0

    K = (n + 1.0) * 6 * E * Ix / L

    McMy = 1.1
    if type_ == 1:
        # RBS hinge
        theta_p = 0.19 * ((htw) ** -0.314) * ((bftf) ** -0.100) * ((Lb / ry) ** -0.185) * ((Ls / d) ** 0.113) * (
                    (c1 * d / 533) ** -0.760) * ((c2 * fy / 355) ** -0.070)
        theta_pc = 9.52 * ((htw) ** -0.513) * ((bftf) ** -0.863) * ((Lb / ry) ** -0.108) * ((c2 * fy / 355) ** -0.360)
        Lamda = 585 * ((htw) ** -1.140) * ((bftf) ** -0.632) * ((Lb / ry) ** -0.205) * ((c2 * fy / 355) ** -0.391)
    elif type_ == 2:
        # Other than RBS section
        if d >= 533.0:
            theta_p = 0.318 * ((htw) ** -0.550) * ((bftf) ** -0.345) * ((Lb / ry) ** -0.023) * ((Ls / d) ** 0.090) * (
                        (c1 * d / 533) ** -0.330) * ((c2 * fy / 355) ** -0.130)
            theta_pc = 7.500 * ((htw) ** -0.610) * ((bftf) ** -0.710) * ((Lb / ry) ** -0.110) * ((c1 * d / 533) ** -0.161) * (
                        (c2 * fy / 355) ** -0.320)
            Lamda = 536 * ((htw) ** -1.260) * ((bftf) ** -0.525) * ((Lb / ry) ** -0.130) * ((c2 * fy / 355) ** -0.291)
        else:
            theta_p = 0.0865 * ((htw) ** -0.360) * ((bftf) ** -0.140) * ((Ls / d) ** 0.340) * ((c1 * d / 533) ** -0.721) * (
                        (c2 * fy / 355) ** -0.230)
            theta_pc = 5.6300 * ((htw) ** -0.565) * ((bftf) ** -0.800) * ((c1 * d / 533) ** -0.280) * (
                        (c2 * fy / 355) ** -0.430)
            Lamda = 495 * ((htw) ** -1.340) * ((bftf) ** -0.595) * ((c2 * fy / 355) ** -0.360)

    if type_ == 3:
        # Beam column hinged connection
        ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", 99, 99, 9, "-dir", 1, 2, 6)
        # ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", 9, "-dir", 6)
        # ops.equalDOF(NodeI, NodeJ, 1, 2)
    elif type_ == 4:
        # Elastic
        ops.uniaxialMaterial("Elastic", SpringID, K)
        ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", 99, 99, SpringID, "-dir", 1, 2, 6)
    else:
        theta_u = 0.2
        D = 1.0
        Res = 0.4
        c = 1.0
        ops.uniaxialMaterial("IMKBilin", SpringID, K, theta_p, theta_pc, theta_u, My, McMy, Res, theta_p, theta_pc, theta_u, My, McMy, Res, Lamda, Lamda, Lamda, c, c, c, D, D)
        ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", 99, 99, SpringID, "-dir", 1, 2, 6)
        # ops.element("zeroLength", SpringID, NodeI, NodeJ, "-mat", SpringID, "-dir", 6)
        # ops.equalDOF(NodeI, NodeJ, 1, 2)

        if check:
            print(f"{check}:\nKs: {K}, My: {My}, theta_p: {theta_p}, theta_pc: {theta_pc}, Res: {Res}")




from . import func
from typing import Dict, Tuple, List, Literal


class UserCommand:
    buildin_mat = [9, 99]

    def __init__(self) -> None:
        self.additional_commands_py: List[Dict[Literal['node', 'ele', 'mat', 'any'], str]] = []
        self.additional_commands_tcl: List[Dict[Literal['node', 'ele', 'mat', 'any'], str]] = []


    def add_material(self, matType: str, Id: int, *paras):
        """Add material

        Args:
            matType (str): Material type
            Id (int): Matirial tag
            paras (tuple[any]): Material parameters
        """
        func.check_int(Id, [1, 10000], name='Id')
        func.check_string(matType, name='matType')
        if Id in self.buildin_mat:
            raise ValueError(f'Material tag {Id} cannot be used since it has been used as a built-in material tag')
        text_paras_tcl = ' '.join([str(i) for i in paras])
        text_tcl = f'uniaxialMaterial {matType} {Id} {text_paras_tcl};'
        self.additional_commands_tcl.append({'mat': text_tcl})
        paras_temp = []
        for item in paras:
            if isinstance(item, str):
                paras_temp.append(f'"{item}"')
            else:
                paras_temp.append(item)
        text_paras_py = ', '.join([str(i) for i in paras_temp])
        text_py = f'ops.uniaxialMaterial("{matType}", {Id}, {text_paras_py})'
        self.additional_commands_py.append({'mat': text_py})
        

    def add_node(self, Id: int, x: int | float, y: int | float):
        """Add node

        Args:
            Id (int): Node tag
            x (int | float): x coordinate
            y (int | float): y coordinate
        """
        func.check_int(Id, [1, 10000], name='Id')
        func.check_int_float(x, pos=False, name='x')
        func.check_int_float(y, pos=False, name='y')
        text_tcl = f'node {Id} {x} {y};'
        self.additional_commands_tcl.append({'node': text_tcl})
        text_py = f'ops.node({Id}, {x}, {y})'
        self.additional_commands_py.append({'node': text_py})


    def add_element(self, eleType: str, Id: int, inode: int, jnode: int, *paras):
        """Add element

        Args:
            eleType (str): Element type
            Id (int): Element tag
            paras (tuple[any]): Element parameters
        """
        func.check_string(eleType, name='eleType')
        func.check_int(Id, [1, 10000], name='Id')
        func.check_int(inode, name='inode')
        func.check_int(jnode, name='jnode')
        text_paras_tcl = ' '.join([str(i) for i in paras])
        text_tcl = f'element {eleType} {Id} {inode} {jnode} {text_paras_tcl};'
        self.additional_commands_tcl.append({'ele': text_tcl})
        paras_temp = []
        for item in paras:
            if isinstance(item, str):
                paras_temp.append(f'"{item}"')
            else:
                paras_temp.append(item)
        text_paras_py = ', '.join([str(i) for i in paras_temp])
        text_py = f'ops.element("{eleType}", {Id}, {inode}, {jnode}, {text_paras_py})'
        self.additional_commands_py.append({'ele': text_py})


    def add_tcl_command(self, command: str=None):
        """Add any command line

        Args:
            command (str): command
        """
        if command is None:
            return
        func.check_string(command, isNone=True, name='command')
        self.additional_commands_tcl.append({'any': command})


    def add_py_command(self, command: str=None):
        """Add any command line

        Args:
            command (str): command
        """
        if command is None:
            return
        func.check_string(command, isNone=True, name='command')
        self.additional_commands_py.append({'any': command})



# -------------------------- Cyclic pushover analysis --------------------------
#
# Args (7):
# ---------------
# CtrlNode          Control node at roof
# RDR_path          Loading path of roof drift ratio
# HBuilding         Heights of the building
# Dincr_init        Initial displacement increment
# maxRunTime        Max run time (s)
# ShowAnimation     Show animation
# min_factor        Factor to control the adaptive time step
# max_factor        Factor to control the adaptive time step

# ---------------
# Written by: Wenchen Lie, Guangzhou University, China
# Date: 2024-05-06
# --------------------------------------------------------------------------------


import numpy as np
import openseespy.opensees as ops
import time


def CyclicPushover(
        CtrlNode: int,
        RDR_path: list,
        HBuilding: float,
        Dincr_init: float,
        maxRunTime: float,
        ShowAnimation: bool,
        min_factor: float=1e-6,
        max_factor: float=1):

    ops.wipeAnalysis()
    ops.constraints("Plain")
    ops.numberer("RCM")
    ops.system("UmfPack")
    ops.test("EnergyIncr", 1.e-5, 30)
    ops.algorithm("KrylovNewton")
    ops.integrator("DisplacementControl", CtrlNode, 1 ,Dincr_init)
    ops.analysis("Static")

    for RDR in RDR_path:

        print(f'RDR = {RDR} starts')
        ls_algorithm = ["KrylovNewton", "NewtonLineSearch", "Newton", "SecantNewton"]
        Id_algorithm = 0
        factor = 1.0
        start_time = time.time()
        D_target = RDR * HBuilding
        if D_target >= ops.nodeDisp(CtrlNode, 1):
            direction = 1
        else:
            direction = -1
        Dincr = direction * Dincr_init
        while True:
            if time.time() - start_time >= maxRunTime:
                print("Exceeding maximum running time")
                return 3
            ops.algorithm(ls_algorithm[Id_algorithm])
            ops.integrator("DisplacementControl", CtrlNode, 1, Dincr)
            ok = ops.analyze(1)
            if ok == 0:
                if (direction == 1) and (ops.nodeDisp(CtrlNode, 1) >= D_target):
                    break
                if (direction == -1) and (ops.nodeDisp(CtrlNode, 1) <= D_target):
                    break
                factor_old = factor
                factor = min(factor * 2, max_factor)
                if factor_old < factor:
                    print(f"-- {ops.nodeDisp(CtrlNode, 1)} -- Enlarged factor: {factor}")
                Id_algorithm -= 1
                Id_algorithm = max(0, Id_algorithm)
            else:
                factor = factor * 0.5
                if factor < min_factor:
                    factor = min_factor
                    Id_algorithm += 1
                    if Id_algorithm == 4:
                        print("Cannot converge")
                        return 2
                    print(f"-- {ops.nodeDisp(CtrlNode, 1)} ------ Switched algorithm: {ls_algorithm[Id_algorithm]}")
                print(f"-- {ops.nodeDisp(CtrlNode, 1)} -- Reduced factor: {factor}")
            Dincr = direction * factor * Dincr_init
        # Analysis finished
        print(f"RDR = {RDR} finished")

    return 1

